# -*- coding: utf-8 -*-

import unittest

from odata.service import ODataService
from odata.entity import declarative_base
from odata.property import StringProperty, IntegerProperty


Base = declarative_base()

NorthwindService = ODataService('http://services.odata.org/V4/Northwind/Northwind.svc/', Base)
service = NorthwindService


class Customer(Base):
    __odata_collection__ = 'Customers'
    __odata_type__ = 'NorthwindModel.Customer'

    id = StringProperty('CustomerID', primary_key=True)
    name = StringProperty('CompanyName')
    contact_name = StringProperty('ContactName')
    contact_title = StringProperty('ContactTitle')
    address = StringProperty('Address')
    city = StringProperty('City')
    region = StringProperty('Region')
    postal_code = StringProperty('PostalCode')
    country = StringProperty('Country')
    phone = StringProperty('Phone')
    fax = StringProperty('Fax')


class Product(Base):
    __odata_collection__ = 'Products'
    __odata_type__ = 'NorthwindModel.Product'

    id = IntegerProperty('ProductID', primary_key=True)
    name = StringProperty('ProductName')
    supplier_id = IntegerProperty('SupplierID')
    category_id = IntegerProperty('CategoryID')
    quantity_per_unit = StringProperty('QuantityPerUnit')


@unittest.skip('unavailable')
class NorthwindManualModelReadTest(unittest.TestCase):

    def test_query_one(self):
        q = service.query(Customer)
        q = q.filter(Customer.contact_title.startswith('Sales'))
        q = q.filter(Customer.postal_code == '68306')
        data = q.first()
        assert data is not None, 'data is None'
        assert isinstance(data, Customer), 'Did not return Customer instance'
        assert data.postal_code == '68306'

    def test_query_all(self):
        q = service.query(Customer)
        q = q.filter(Customer.city != 'Berlin')
        q = q.limit(30)
        q = q.order_by(Customer.city.asc())
        data = q.all()
        assert data is not None, 'data is None'
        assert len(data) > 20, 'data length wrong'

    def test_iterating_query_result(self):
        q = service.query(Customer)
        q = q.limit(20)
        for result in q:
            assert isinstance(result, Customer), 'Did not return Customer instance'

    def test_query_raw_data(self):
        q = service.query(Customer)
        q = q.select(Customer.name)
        data = q.first()
        assert isinstance(data, dict), 'Did not return dict'
        assert Customer.name.name in data

    def test_query_filters(self):
        q = service.query(Product)

        q = q.filter(
            q.or_(
                q.grouped(
                    q.or_(
                        Product.name.startswith('Chai'),
                        Product.name.startswith('Chang'),
                    )
                ),
                Product.quantity_per_unit == '12 - 550 ml bottles',
            )
        )

        data = q.all()
        assert len(data) == 3, 'data length wrong'



# -*- coding: utf-8 -*-

import unittest
import json
import decimal

import responses
import requests

from odata.tests import Product, Service, DemoUnboundAction


class TestActions(unittest.TestCase):

    def test_call_bound_action_empty_result(self):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST, Product.__odata_url__() + '/ODataTest.DemoCollectionAction',
            )
            result = Product.DemoCollectionAction()
        self.assertIsNone(result)

    def test_call_bound_action_on_instance(self):
        test_product = Product()
        test_product.name = 'TestProduct'
        test_product.id = 1234
        test_product.price = decimal.Decimal('20.0')

        # shortcut for saving the entity
        test_product.__odata__.persisted = True

        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST, test_product.__odata__.instance_url + '/ODataTest.DemoAction',
            )
            result = test_product.DemoAction()
        self.assertIsNone(result)

    def test_call_instance_bound_action_on_collection(self):
        def _call():
            Product.DemoAction()
        self.assertRaises(AttributeError, _call)

    def test_call_bound_action_with_result(self):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST, Product.__odata_url__() + '/ODataTest.DemoCollectionAction',
                content_type='application/json',
                json=dict(value='test'),
            )
            result = Product.DemoCollectionAction()
        self.assertEqual(result, 'test')

    def test_call_unbound_action_with_result(self):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST, Service.url + 'ODataTest.DemoUnboundAction',
                content_type='application/json',
                json=dict(value='test'),
            )
            result = DemoUnboundAction()
        self.assertEqual(result, 'test')


class TestActionsWithParameters(unittest.TestCase):

    def test_call_action_with_result(self):
        def request_callback(request):
            payload = json.loads(request.body)
            expected = dict(Name='TestName', Price=decimal.Decimal('10.0'))

            self.assertDictEqual(payload, expected)

            headers = {}
            body = dict(value='ok')
            return requests.codes.ok, headers, json.dumps(body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.POST, Product.__odata_url__() + '/ODataTest.DemoActionParameters',
                callback=request_callback, content_type='application/json',
            )
            result = Product.DemoActionWithParameters(Name='TestName', Price=decimal.Decimal('10.0'))
        self.assertEqual(result, 'ok')


class TestFunctions(unittest.TestCase):

    def test_call_function_empty_result(self):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.GET, Product.__odata_url__() + '/ODataTest.DemoFunction()',
                content_type='application/json', status=requests.codes.no_content
            )
            result = Product.DemoFunction()
        self.assertIsNone(result)

    def test_collection_bound_function_on_instance(self):
        test_product = Product()

        def _call():
            test_product.DemoFunction()

        self.assertRaises(AttributeError, _call)

    def test_call_function_with_result_query(self):
        def request_callback(request):
            self.assertTrue('filter=ProductName+eq+%27testtest%27' in request.url)

            headers = {}
            body = dict(value='ok')
            return requests.codes.ok, headers, json.dumps(body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, Product.__odata_url__() + '/ODataTest.DemoFunction()',
                request_callback, content_type='application/json')

            query = Service.query(Product)
            query = query.filter(Product.name == 'testtest')
            Product.DemoFunction.with_query(query)()



# -*- coding: utf-8 -*-

from odata.property import PropertyBase


class ComplexType(dict):
    properties = dict()

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __repr__(self):
        keys = ','.join(self.keys())
        return '<ComplexType({0})>'.format(keys)


class ComplexTypeProperty(PropertyBase):
    """
    A property that contains a ComplexType object

    :param name: Name of the property in the endpoint
    :param type_class: A subclass of ComplexType
    """

    def __init__(self, name, type_class=ComplexType):
        """
        :type name: str
        """
        super(ComplexTypeProperty, self).__init__(name)
        self.type_class = type_class

    def serialize(self, value):
        if isinstance(value, list):
            data = []
            for i in value:
                data.append(self._serialize(i))
            return data
        else:
            return self._serialize(value)

    def _serialize(self, value):
        data = dict()
        for name, prop in value.properties.items():
            prop_value = value.get(name)

            if prop_value is None:
                continue

            if isinstance(prop_value, ComplexType):
                serialized_value = self.serialize(prop_value)
            else:
                serialized_value = prop('temp').serialize(prop_value)
            data[name] = serialized_value
        return data

    def deserialize(self, value):
        if isinstance(value, list):
            data = []
            for i in value:
                data.append(self._deserialize(i))
            return data
        else:
            return self._deserialize(value)

    def _deserialize(self, value):
        data = self.type_class()

        for name, prop in data.properties.items():
            prop_value = value.get(name)

            if prop_value is None:
                continue

            if issubclass(prop, ComplexType):
                ctprop = ComplexTypeProperty('temp', type_class=prop)
                deserialized_value = ctprop.deserialize(prop_value)
            else:
                deserialized_value = prop('temp').deserialize(prop_value)
            data[name] = deserialized_value
        return data

    def escape_value(self, value):
        raise NotImplementedError()

    def __getattr__(self, item):
        # allows ComplexType key usage in filters etc
        subkey = '{0}/{1}'.format(self.name, item)
        prop = self.type_class.properties[item]
        if issubclass(prop, ComplexType):
            return ComplexTypeProperty(subkey, type_class=prop)
        return prop(subkey)



import time
from ..api.client import Client

api = Client()

def call(source, number):
    """Call number form source"""
    name = source.Name
    print("Calling {0} from {1}".format(number, name))
    return api.post("cti/calls/{0}/{1}".format(name, number), None)

def answer(source):
    print("Answering call on {0}".format(source.Name))
    return operation(source,'Answer')

def drop(source):
    print("Dropping call on {0}".format(source.Name))
    return operation(source,'Drop')

def hold(source):
    """Place the active call on the source phone on hold"""
    print("Holding call on {0}".format(source.Name))
    return operation(source,'Hold')

def unhold(source):
    print("Unholding call on {0}".format(source.Name))
    return operation(source,'UnHold')

def transfer(source):
    print("Consult Transfer on {0}".format(source.Name))
    return operation(source,'Transfer')

def conference(source):
    print("Consult Conference on {0}".format(source.Name))
    return operation(source,'Conference')

def offhook(source):
    print("Offhook on {0}".format(source.Name))
    return operation(source,'OffHook')

def operation(source, op):
    name = source.Name
    return api.post("cti/calls/{0}/{1}".format(name,op), None)


def sendDigits(source, digits):
    name = source.Name
    print("Send Digits '{0}' on {1}".format(digits, name))
    return api.post("cti/calls/{0}/digits/{1}".format(name,digits), None)

def sendData(source, data):
    name = source.Name
    return api.post("cti/data/{0}".format(name), data=data)

def sendUri(source, uri):
    name = source.Name
    return api.postString("cti/execute/{0}".format(name), uri)

def pause(delay):
    print("Pausing for {0} seconds".format(delay))
    time.sleep(delay)

def macro(source, macro):
    name = source.Name
    print("Sending macro '{0}' to {1}".format(macro, name))
    return api.postString("phones/macro/{0}".format(name),macro)

def sendMessage(message, destination):
    """
    Send Message
    ============
    Send a text/audio message to a group of phone using an existing message and filter defined in the AutomationFX admin interface
    - message (the name of the message to send)
    - destination (the name of the filter to send the message too)
    """
    return api.post("messages/{0}/send".format(message), None, params={'filter':destination})

def sendRawMessage(message, source):
    return api.post("messages/send", data=message, params={'filter':source})

def listResource(resource):
    return api.get("axl/{0}".format(resource), params={'detailed':'true'})

def getResource(resource, id):
    return api.get("axl/{0}/{1}".format(resource,id))

def addResource(resource, data):
    return api.post("axl/{0}".format(resource), data=data)

def updateResource(resource, id, data):
    return api.put("axl/{0}/{1}".format(resource,id),data)

def deleteResource(resource, id):
    return api.delete("axl/{0}/{1}".format(resource,id))

def sqlQuery(query):
    return api.postString("axl/sql/query", query)

def sqlUpdate(update):
    return api.postString("axl/sql/update", update)



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import inspect
from collections import OrderedDict

from .property import PropertyBase, NavigationProperty


class EntityState(object):

    def __init__(self, entity):
        """:type entity: EntityBase """
        self.entity = entity
        self.dirty = []
        self.nav_cache = {}
        self.data = {}
        self.connection = None
        # does this object exist serverside
        self.persisted = False

    # dictionary access
    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data

    def get(self, key, default):
        return self.data.get(key, default=default)

    def update(self, other):
        self.data.update(other)
    # /dictionary access

    def __repr__(self):
        return self.data.__repr__()

    def describe(self):
        rows = [
            u'EntitySet: {0}'.format(self.entity.__odata_collection__),
            u'Type: {0}'.format(self.entity.__odata_type__),
            u'URL: {0}'.format(self.instance_url or self.entity.__odata_url__()),
            u'',
            u'Properties',
            u'-' * 40,
        ]

        for _, prop in self.properties:
            name = prop.name
            if prop.primary_key:
                name += '*'
            if prop.name in self.dirty:
                name += ' (dirty)'
            rows.append(name)

        rows.append(u'')
        rows.append(u'Navigation Properties')
        rows.append(u'-' * 40)

        for _, prop in self.navigation_properties:
            rows.append(prop.name)

        rows = os.linesep.join(rows)
        print(rows)

    def reset(self):
        self.dirty = []
        self.nav_cache = {}

    @property
    def id(self):
        ids = []
        entity_name = self.entity.__odata_collection__
        for prop_name, prop in self.primary_key_properties:
            value = self.data.get(prop.name)
            if value:
                ids.append((prop, str(prop.escape_value(value))))
        if len(ids) == 1:
            key_value = ids[0][1]
            return u'{0}({1})'.format(entity_name,
                                      key_value)
        if len(ids) > 1:
            key_ids = []
            for prop, key_value in ids:
                key_ids.append('{0}={1}'.format(prop.name, key_value))
            return u'{0}({1})'.format(entity_name, ','.join(key_ids))

    @property
    def instance_url(self):
        if self.id:
            return self.entity.__odata_url_base__ + self.id

    @property
    def properties(self):
        props = []
        cls = self.entity.__class__
        for key, value in inspect.getmembers(cls):
            if isinstance(value, PropertyBase):
                props.append((key, value))
        return props

    @property
    def primary_key_properties(self):
        pks = []
        for prop_name, prop in self.properties:
            if prop.primary_key is True:
                pks.append((prop_name, prop))
        return pks

    @property
    def navigation_properties(self):
        props = []
        cls = self.entity.__class__
        for key, value in inspect.getmembers(cls):
            if isinstance(value, NavigationProperty):
                props.append((key, value))
        return props

    @property
    def dirty_properties(self):
        rv = []
        for prop_name, prop in self.properties:
            if prop.name in self.dirty:
                rv.append((prop_name, prop))
        return rv

    def set_property_dirty(self, prop):
        if prop.name not in self.dirty:
            self.dirty.append(prop.name)

    def data_for_insert(self):
        return self._clean_new_entity(self.entity)

    def data_for_update(self):
        update_data = OrderedDict()
        update_data['@odata.type'] = self.entity.__odata_type__

        for _, prop in self.dirty_properties:
            if prop.is_computed_value:
                continue

            update_data[prop.name] = self.data[prop.name]

        for prop_name, prop in self.navigation_properties:
            if prop.name in self.dirty:
                value = getattr(self.entity, prop_name, None)  # get the related object
                """:type : None | odata.entity.EntityBase | list[odata.entity.EntityBase]"""
                if value is not None:
                    key = '{0}@odata.bind'.format(prop.name)
                    if prop.is_collection:
                        update_data[key] = [i.__odata__.id for i in value]
                    else:
                        update_data[key] = value.__odata__.id
        return update_data

    def _clean_new_entity(self, entity):
        """:type entity: odata.entity.EntityBase """
        insert_data = OrderedDict()
        insert_data['@odata.type'] = entity.__odata_type__

        es = entity.__odata__
        for _, prop in es.properties:
            if prop.is_computed_value:
                continue

            insert_data[prop.name] = es[prop.name]

        # Allow pk properties only if they have values
        for _, pk_prop in es.primary_key_properties:
            if insert_data[pk_prop.name] is None:
                insert_data.pop(pk_prop.name)

        # Deep insert from nav properties
        for prop_name, prop in es.navigation_properties:
            if prop.foreign_key:
                insert_data.pop(prop.foreign_key, None)

            value = getattr(entity, prop_name, None)
            """:type : None | odata.entity.EntityBase | list[odata.entity.EntityBase]"""
            if value is not None:

                if prop.is_collection:
                    binds = []

                    # binds must be added first
                    for i in [i for i in value if i.__odata__.id]:
                        binds.append(i.__odata__.id)

                    if len(binds):
                        insert_data['{0}@odata.bind'.format(prop.name)] = binds

                    new_entities = []
                    for i in [i for i in value if i.__odata__.id is None]:
                        new_entities.append(self._clean_new_entity(i))

                    if len(new_entities):
                        insert_data[prop.name] = new_entities

                else:
                    if value.__odata__.id:
                        insert_data['{0}@odata.bind'.format(prop.name)] = value.__odata__.id
                    else:
                        insert_data[prop.name] = self._clean_new_entity(value)

        return insert_data



# -*- coding: utf-8 -*-

"""
Navigation properties
---------------------

The entity can define properties that link to other entities. These are known
as navigation properties and are supported in this library.

.. code-block:: python

    >>> order = Service.query(Order).first()
    >>> order.Shipper
    <Entity(Shipper:3)>
    >>> order.Shipper.CompanyName
    'Federal Shipping'

When creating new instances, relationships can be assigned via navigation
properties:

.. code-block:: python

    # query a shipper instance, just for this example
    Shipper = Service.entities['Shipper']
    my_shipper = Service.query(Shipper).first()

    # assign for the new Order
    order.Shipper = my_shipper
    Service.save(order)
"""

try:
    # noinspection PyUnresolvedReferences
    from urllib.parse import urljoin
except ImportError:
    # noinspection PyUnresolvedReferences
    from urlparse import urljoin


class NavigationProperty(object):
    """
    A Property-like object for marking relationships between entities, but does
    not inherit from PropertyBase.
    """
    def __init__(self, name, entitycls, collection=False, foreign_key=None):
        from .property import PropertyBase
        self.name = name
        self.entitycls = entitycls
        self.is_collection = collection
        if isinstance(foreign_key, PropertyBase):
            self.foreign_key = foreign_key.name
        else:
            self.foreign_key = foreign_key

    def __repr__(self):
        return u'<NavigationProperty to {0}>'.format(self.entitycls)

    def instances_from_data(self, raw_data):
        if self.is_collection:
            return [self.entitycls.__new__(self.entitycls, from_data=d) for d in raw_data]
        else:
            return self.entitycls.__new__(self.entitycls, from_data=raw_data)

    def _get_parent_cache(self, instance):
        es = instance.__odata__
        ic = es.nav_cache
        if self.name not in ic:
            cache = {}
            ic[self.name] = cache
        else:
            cache = ic[self.name]
        return cache

    def __set__(self, instance, value):
        """
        :type instance: odata.entity.EntityBase
        """
        cache = self._get_parent_cache(instance)
        if self.is_collection:
            cache['collection'] = value
        else:
            cache['single'] = value
        instance.__odata__.set_property_dirty(self)

    def __get__(self, instance, owner):
        """
        :type instance: odata.entity.EntityBase
        """
        if instance is None:
            return self

        es = instance.__odata__
        connection = es.connection
        parent_url = es.instance_url
        new_object = parent_url is None
        cache = self._get_parent_cache(instance)

        if new_object:
            if self.is_collection:
                return cache.get('collection', [])
            return cache.get('single', None)

        parent_url += '/'
        url = urljoin(parent_url, self.name)

        if self.is_collection:
            if 'collection' not in cache:
                raw_data = connection.execute_get(url)
                if raw_data:
                    cache['collection'] = self.instances_from_data(raw_data['value'])
                else:
                    cache['collection'] = []
            return cache['collection']
        else:
            if 'single' not in cache:
                raw_data = connection.execute_get(url)
                if raw_data:
                    cache['single'] = self.instances_from_data(raw_data)
                else:
                    cache['single'] = None
            return cache['single']



# -*- coding: utf-8 -*-

import unittest

from odata.service import ODataService


url = 'http://services.odata.org/V4/Northwind/Northwind.svc/'
Service = ODataService(url, reflect_entities=True)
Customer = Service.entities.get('Customer')
Product = Service.entities.get('Product')


@unittest.skip('unavailable')
class NorthwindReflectModelReadTest(unittest.TestCase):

    def test_query_one(self):
        q = Service.query(Customer)
        q = q.filter(Customer.ContactTitle.startswith('Sales'))
        q = q.filter(Customer.PostalCode == '68306')
        data = q.first()
        assert data is not None, 'data is None'
        assert isinstance(data, Customer), 'Did not return Customer instance'
        assert data.PostalCode == '68306'

    def test_query_all(self):
        q = Service.query(Customer)
        q = q.filter(Customer.City != 'Berlin')
        q = q.limit(30)
        q = q.order_by(Customer.City.asc())
        data = q.all()
        assert data is not None, 'data is None'
        assert len(data) > 20, 'data length wrong'

    def test_iterating_query_result(self):
        q = Service.query(Customer)
        q = q.limit(20)
        for result in q:
            assert isinstance(result, Customer), 'Did not return Customer instance'

    def test_query_raw_data(self):
        q = Service.query(Customer)
        q = q.select(Customer.CompanyName)
        data = q.first()
        assert isinstance(data, dict), 'Did not return dict'
        assert Customer.CompanyName.name in data

    def test_query_filters(self):
        q = Service.query(Product)

        q = q.filter(
            q.or_(
                q.grouped(
                    q.or_(
                        Product.ProductName.startswith('Chai'),
                        Product.ProductName.startswith('Chang'),
                    )
                ),
                Product.QuantityPerUnit == '12 - 550 ml bottles',
            )
        )

        data = q.all()
        assert len(data) == 3, 'data length wrong'



# -*- coding: utf-8 -*-

from enum import Enum

from .property import PropertyBase


class EnumType(Enum):
    pass


class EnumTypeProperty(PropertyBase):
    """
    A property that contains a ComplexType object

    :param name: Name of the property in the endpoint
    :param enum_class: A subclass of EnumType
    """

    def __init__(self, name, enum_class=EnumType):
        super(EnumTypeProperty, self).__init__(name)
        self.enum_class = enum_class

    def serialize(self, value):
        return value.name

    def deserialize(self, value):
        return self.enum_class[value]



# -*- coding: utf-8 -*-

import json
import base64
import decimal
from unittest import TestCase

import requests
import responses

from odata.tests import Service, Product, DemoUnboundAction


class TestContext(TestCase):

    def test_context_query_without_auth(self):
        def request_callback(request):
            self.assertIsNone(request.headers.get('Authorization'))
            headers = {}
            body = dict(value=[])
            return requests.codes.ok, headers, json.dumps(body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(rsps.GET, Product.__odata_url__(),
                              callback=request_callback,
                              content_type='application/json')

            context = Service.create_context()
            context.query(Product).first()

    def test_context_query_with_basic_auth(self):
        test_username = 'username'
        test_password = 'password'
        test_auth = (test_username, test_password)

        def request_callback(request):
            auth_text = request.headers.get('Authorization')
            _, auth_b64 = auth_text.split(' ', 1)
            decoded = base64.urlsafe_b64decode(auth_b64.encode()).decode()
            username, password = decoded.split(':', 1)

            self.assertEqual(test_username, username)
            self.assertEqual(test_password, password)

            headers = {}
            body = dict(value=[])
            return requests.codes.ok, headers, json.dumps(body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(rsps.GET, Product.__odata_url__(), request_callback,
                              content_type='application/json')

            context = Service.create_context(auth=test_auth)
            context.query(Product).first()

    def test_context_call_unbound_action(self):
        with responses.RequestsMock() as rsps:
            rsps.add(rsps.POST, Service.url + 'ODataTest.DemoUnboundAction')

            context = Service.create_context()
            context.call(DemoUnboundAction)

    def test_context_call_bound_action(self):
        with responses.RequestsMock() as rsps:
            rsps.add(rsps.POST, Product.__odata_url__() + '/ODataTest.DemoActionParameters')

            context = Service.create_context()
            context.call(Product.DemoActionWithParameters,
                         Name='TestName',
                         Price=decimal.Decimal('25.0'))



# -*- coding: utf-8 -*-

import json
import functools
import logging

import requests
from requests.exceptions import RequestException

from automationfx.odata import version
from .exceptions import ODataError, ODataConnectionError


def catch_requests_errors(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except RequestException as e:
            raise ODataConnectionError(str(e))
    return inner


class ODataConnection(object):

    base_headers = {
        'Accept': 'application/json',
        'OData-Version': '4.0',
        'User-Agent': 'python-odata {0}'.format(version),
    }
    timeout = 90

    def __init__(self, session=None, auth=None):
        if session is None:
            self.session = requests.Session()
        else:
            self.session = session
        self.auth = auth
        self.log = logging.getLogger('odata.connection')

    def _apply_options(self, kwargs):
        kwargs['timeout'] = self.timeout

        if self.auth is not None:
            kwargs['auth'] = self.auth

    @catch_requests_errors
    def _do_get(self, *args, **kwargs):
        self._apply_options(kwargs)
        return self.session.get(*args, **kwargs)

    @catch_requests_errors
    def _do_post(self, *args, **kwargs):
        self._apply_options(kwargs)
        return self.session.post(*args, **kwargs)

    @catch_requests_errors
    def _do_patch(self, *args, **kwargs):
        self._apply_options(kwargs)
        return self.session.patch(*args, **kwargs)

    @catch_requests_errors
    def _do_delete(self, *args, **kwargs):
        self._apply_options(kwargs)
        return self.session.delete(*args, **kwargs)

    def _handle_odata_error(self, response):
        try:
            response.raise_for_status()
        except:
            status_code = 'HTTP {0}'.format(response.status_code)
            code = 'None'
            message = 'Server did not supply any error messages'
            detailed_message = 'None'
            response_ct = response.headers.get('content-type', '')

            if 'application/json' in response_ct:
                errordata = response.json()

                if 'error' in errordata:
                    odata_error = errordata.get('error')

                    if 'code' in odata_error:
                        code = odata_error.get('code') or code
                    if 'message' in odata_error:
                        message = odata_error.get('message') or message
                    if 'innererror' in odata_error:
                        ie = odata_error['innererror']
                        detailed_message = ie.get('message') or detailed_message

            msg = ' | '.join([status_code, code, message, detailed_message])
            err = ODataError(msg)
            err.status_code = status_code
            err.code = code
            err.message = message
            err.detailed_message = detailed_message
            raise err

    def execute_get(self, url, params=None):
        headers = {}
        headers.update(self.base_headers)

        self.log.info(u'GET {0}'.format(url))
        if params:
            self.log.info(u'Query: {0}'.format(params))

        response = self._do_get(url, params=params, headers=headers)
        self._handle_odata_error(response)
        response_ct = response.headers.get('content-type', '')
        if response.status_code == requests.codes.no_content:
            return
        if 'application/json' in response_ct:
            data = response.json()
            return data
        else:
            msg = u'Unsupported response Content-Type: {0}'.format(response_ct)
            raise ODataError(msg)

    def execute_post(self, url, data, params=None):
        headers = {
            'Content-Type': 'application/json',
        }
        headers.update(self.base_headers)

        data = json.dumps(data)

        self.log.info(u'POST {0}'.format(url))
        self.log.info(u'Payload: {0}'.format(data))

        response = self._do_post(url, data=data, headers=headers, params=params)
        self._handle_odata_error(response)
        response_ct = response.headers.get('content-type', '')
        if response.status_code == requests.codes.no_content:
            return
        if 'application/json' in response_ct:
            return response.json()
        # no exceptions here, POSTing to Actions may not return data

    def execute_patch(self, url, data):
        headers = {
            'Content-Type': 'application/json',
        }
        headers.update(self.base_headers)

        data = json.dumps(data)

        self.log.info(u'PATCH {0}'.format(url))
        self.log.info(u'Payload: {0}'.format(data))

        response = self._do_patch(url, data=data, headers=headers)
        self._handle_odata_error(response)

    def execute_delete(self, url):
        headers = {}
        headers.update(self.base_headers)

        self.log.info(u'DELETE {0}'.format(url))

        response = self._do_delete(url, headers=headers)
        self._handle_odata_error(response)



# -*- coding: utf-8 -*-

from decimal import Decimal
from unittest import TestCase
import json

import requests
import responses

from odata.tests import Service, ProductManufacturerSales


class TestCompositeKeys(TestCase):

    def test_insert_entity(self):
        def request_callback(request):
            content = json.loads(request.body)
            self.assertIn('ProductID', content)
            self.assertIn('ManufacturerID', content)
            self.assertIsNotNone(content.get('ProductID'),
                                 msg='ProductID received as None')
            self.assertIsNotNone(content.get('ManufacturerID'),
                                 msg='ManufacturerID received as None')

            headers = {}
            return requests.codes.ok, headers, json.dumps(content)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(rsps.POST,
                              ProductManufacturerSales.__odata_url__(),
                              callback=request_callback,
                              content_type='application/json')

            pm_sales = ProductManufacturerSales()
            pm_sales.product_id = 1
            pm_sales.manufacturer_id = 2
            pm_sales.sales_amount = Decimal('3.0')
            Service.save(pm_sales)

    def test_update_entity(self):
        test_pm_sales_value = dict(
            ProductID=1,
            ManufacturerID=2,
            SalesAmount=23.0
        )

        with responses.RequestsMock() as rsps:
            rsps.add(rsps.GET, ProductManufacturerSales.__odata_url__(),
                     content_type='application/json',
                     json=dict(value=[test_pm_sales_value]))

            query = Service.query(ProductManufacturerSales)
            query = query.filter(ProductManufacturerSales.product_id == 1)
            query = query.filter(ProductManufacturerSales.manufacturer_id == 2)
            pm_sales = query.first()  # type: ProductManufacturerSales

            sales_id = pm_sales.__odata__.id
            self.assertIn('ProductID=1', sales_id)
            self.assertIn('ManufacturerID=2', sales_id)

            rsps.add(rsps.PATCH, pm_sales.__odata__.instance_url,
                     content_type='application/json')
            rsps.add(rsps.GET, pm_sales.__odata__.instance_url,
                     content_type='application/json',
                     json=dict(value=[test_pm_sales_value]))

            pm_sales.sales_amount = Decimal('50.0')
            Service.save(pm_sales)
